//#include <avr/io.h>

#include <stdlib.h>
#include <stdint.h>

#include <avr/io.h>
#include <avr/interrupt.h>
//#include <avr/signal.h>
#include <avr/pgmspace.h>
#include "uart.h"


//typedef unsigned int uint16_t;
//typedef unsigned char uint8_t;

#ifndef F_CPU
//#define F_CPU  8000000UL
#define F_CPU   14475600UL
#endif

#ifndef UART_BAUD_RATE
#define UART_BAUD_RATE 38400UL
//#define UART_BAUD_RATE 1200UL
//#define UART_BAUD_RATE 9600UL
#endif

#if 1

volatile uint32_t dummy = 0;
volatile uint32_t idleperiod = 0;

//volatile uint8_t quiet = 0;

volatile uint16_t pf_cmd2;

volatile int8_t  pf_repeats = 0;

volatile uint16_t pf_cmd;
volatile uint16_t pf_mask;
volatile uint16_t pf_cmd_prev;
volatile uint16_t pf_cmd_prev;
volatile uint16_t pf_cmd_recv;
volatile uint16_t pf_cmd_recv_valid = 0;
volatile uint16_t period;

volatile uint8_t  nperiods = 0;
volatile uint16_t periods[64];

volatile enum {
  rcv_idle = 0,
	rcv_initial_low,
	rcv_initial_high,
	rcv_second_low,
	rcv_low,
  rcv_high,
	rcv_quiet,
  rcv_done,

  snd_idle,
	snd_initial_low,
	snd_initial_high,
	snd_second_low,
	snd_low,
  snd_high,
	snd_quiet,
  snd_done
} state; 

ISR(TIMER1_COMPA_vect) {
#if 1
  uint16_t xxx = TCNT1;
  TCNT1 = 0;
  switch (state) {
	  case snd_idle:
		  break;
    case snd_initial_low:
		  nperiods = 0;
		  periods[ nperiods++ ] = xxx;
      TCCR0A &= ~_BV(COM0B1); 
    	PORTD &= ~0x20;
      OCR1A = 223; // initial high period
			state = snd_initial_high;
			break;
	  case snd_initial_high:
		  periods[ nperiods++ ] = xxx;
      TCCR0A |= _BV(COM0B1); 
      OCR1A = 60; // low period
		  state = snd_second_low;
		  break;
    case snd_second_low:
    case snd_low:
		  periods[ nperiods++ ] = xxx;
		  // no signal
      TCCR0A &= ~_BV(COM0B1); 
    	PORTD &= ~0x20;
			if (pf_cmd & 0x8000)
        OCR1A = 110; // long period
      else
        OCR1A = 40; // long period
			if (pf_mask)
			  state = snd_high;
      else {
			  state = snd_quiet;
        OCR1A = 15000U; 
		  }
      pf_mask <<= 1;
			pf_cmd  <<= 1;
			break;
	  case snd_high:
		  periods[ nperiods++ ] = xxx;
      TCCR0A |= _BV(COM0B1); 
      OCR1A = 60; // low period
		  state = snd_low;
		  break;
	  case snd_quiet:
		  if (pf_repeats != 0) {
			  if (pf_repeats > 0) pf_repeats--;
				pf_cmd = pf_cmd_prev;
			  pf_send_start();
			} else {
  		  state = rcv_idle;
				idleperiod = 0;
				OCR1A = 768; // set for receive operations
			}
		  break;
	  case snd_done:
		  break;

	  case rcv_idle:
		  idleperiod += 768;
		  break;
	  case rcv_high:
		  pf_cmd_recv = pf_cmd;
			pf_cmd_recv_valid = 1;
		  state = rcv_idle; // was done
		  break;
    default:
		  break;
	}
#endif
}

ISR(TIMER1_CAPT_vect) {
  switch (state) {
	  case rcv_idle:
		  idleperiod += ICR1;
      TCCR1B |= _BV(ICES1); // wait for rising edge
			state = rcv_initial_low;
      TCNT1 = TCNT1 - ICR1;
		  break;
    case rcv_initial_low:
      period = ICR1;
		  if (period >= 50 && period <= 80) {
  			state = rcv_initial_high;
			} else
			  state = rcv_idle;
      TCCR1B &= ~(_BV(ICES1)); // wait for falling edge
      TCNT1 = TCNT1 - ICR1;
		  break;
    case rcv_initial_high:
      period = ICR1;
		  if (period >= 190 && period <= 235) {
  			state = rcv_second_low;
			} else
			  state = rcv_initial_low; // try again ...
      TCCR1B |= _BV(ICES1); // wait for rising edge
      TCNT1 = TCNT1 - ICR1;
		  break;
    case rcv_second_low:
      period = ICR1;
		  if (period >= 50 && period <= 80) {
				pf_cmd = 0;
  			state = rcv_high;
			} else
			  state = rcv_idle;
      TCCR1B &= ~(_BV(ICES1)); // wait for falling edge
      TCNT1 = TCNT1 - ICR1;
		  break;
    case rcv_low:
      period += ICR1;
  		pf_cmd <<= 1;
		  if (period >= 60 && period <= 143) {
  			state = rcv_high;
			} else if (period > 143 && period <= 200) {
			  pf_cmd |= 0x0001;
  			state = rcv_high;
			} else
				state = rcv_idle;
      TCCR1B &= ~(_BV(ICES1)); // wait for falling edge
      TCNT1 = TCNT1 - ICR1;
		  break;
	  case rcv_high:
      period = ICR1;
      TCCR1B |= _BV(ICES1); // wait for rising edge
			state = rcv_low;
      TCNT1 = TCNT1 - ICR1;
		  break;
	  default:
		  break;
	}

}

pf_init() 
{
	cli();

	/*
	 *  Set timer 0 to generate the 38 KHz square wave
	 */

  TCCR0A = 0 //  _BV(COM0B1) // set at bottom, clear on match
	         | _BV(WGM00)  // two bits of fast PWM 
					 | _BV(WGM01);
				
  TCCR0B =   _BV(WGM02) // finish waveform generation
	         | _BV(CS01); // divide by 8; gives about 49 counts per 38KHz cycle

	OCR0A = 48; // period
  OCR0B = 20; // high time within period

	DDRD |= 0x20;   // set pin PD5/OC0B to output
	PORTD &= ~0x20; // set to zero for now.

  // from receive init

 	TCCR1B =  _BV(ICNC1) // noise canceler
	                     // not setting ICES1, so falling edge
         | _BV(CS11) | _BV(CS10) // 011, meaning divide by 64
				 ; 

  TIMSK1 = _BV(ICIE1)   // enable capture interrupt
	       | _BV(TOIE1)   // enable overflow interrupt
	       | _BV(OCIE1A); // enable compare-match-A interrupt


  TCNT1 = 0;

  OCR1A = 768; // maximum high before we declare detection

  // from send-init

  /*
	 * set timer 1 to generate the pulse-distance modulation
	 */

	state = rcv_idle;

	sei();
}

void pf_receive() {
  char buffer[7];
	cli();

 	TCCR1B =  _BV(ICNC1) // noise canceler
	                     // not setting ICES1, so falling edge
         | _BV(CS11) | _BV(CS10) // 011, meaning divide by 64
				 ; 

  TIMSK1 = _BV(ICIE1)   // enable capture interrupt
	       | _BV(TOIE1)   // enable overflow interrupt
	       | _BV(OCIE1A); // enable compare-match-A interrupt


  TCNT1 = 0;

  OCR1A = 768; // maximum high before we declare detection

	sei();

	while (0) {
	  uint8_t print = 0;
		cli();
  	if (state == rcv_done) {
			print = 1;
		}
		sei();
		if (print) {
      uart_putc('*');
			cli();
			state = rcv_idle;
			sei();
    }
	}

  state = rcv_idle;
	while (1) {
	  uint8_t print = 0;
		uint32_t idlep;
		cli();
  	if (state == rcv_done) {
			print = 1;
			idlep = idleperiod;
			idleperiod = 768; // already overflowed once...
		}
		sei();
		if (print) {
		  uint16_t mask = 0x8000;
			while (mask) {
			  if (pf_cmd & mask)
          uart_putc('1');
        else
          uart_putc('0');
        mask >>= 1;
			}
      uart_putc(' ');
      ltoa(idlep, buffer, 10);
      uart_puts(buffer);
      uart_putc('\r');
      uart_putc('\n');
			cli();
			state = rcv_idle;
			sei();
    }
	}
}

void pf_send_init() 
{

  /*
	 * set timer 0 to generate a 38KHz output on pin OC0B
	 */

  TCCR0A = 0 //  _BV(COM0B1) // set at bottom, clear on match
	         | _BV(WGM00)  // two bits of fast PWM 
					 | _BV(WGM01);
				
  TCCR0B =   _BV(WGM02) // finish waveform generation
	         | _BV(CS01); // divide by 8; gives about 49 counts per 38KHz cycle

  TIMSK1 = 0; // no interrupts for now
	
	OCR0A = 48; // period
  OCR0B = 20; // high time within period

  /*
	 * set timer 1 to generate the pulse-distance modulation
	 */

 	TCCR1B =  _BV(CS11) | _BV(CS10); // 011, meaning divide by 64

	DDRD |= 0x20;   // set pin PD5/OC0B to output
	PORTD &= ~0x20; // set to zero for now.

	state = rcv_idle;
}

void pf_send_start()
{
  //TIMSK1 = _BV(ICIE1)  // enable capture interrupt
	//       | _BV(TOIE1); // enable overflow interrupt

	state = snd_initial_low;
	pf_mask = 0xff;
	pf_cmd_prev = pf_cmd;

  OCR1A = 60;
  TCNT1 = 0;

  TCCR0A |= _BV(COM0B1); // set at bottom, clear on match

//  TIMSK1 = _BV(OCIE1A); // enable compare-match-A interrupt
}

/*
 * I2C routines
 */

volatile uint8_t i2c_in_msg[4];
// the out buffer is filled in reverse, with the last byte to be sent
// always in position 0, the previous byte in position 1, and so on.
volatile uint8_t i2c_out_msg[4];

volatile uint8_t i2c_in_msg_len;
volatile uint8_t i2c_out_msg_len;

#define SLAVE_ADDRESS 0x23

volatile uint8_t i2c_processed = 0;

void i2c_process_in_msg()
{
  i2c_processed = 1;
  switch (i2c_in_msg[0]) {
	  case 0x01:
		  pf_repeats = 5; // one shot
		  pf_cmd = i2c_in_msg[1];
			pf_cmd <<= 8;
		  pf_cmd |= i2c_in_msg[2];
			pf_send_start();
		  break;
	  case 0x02:
		  pf_repeats = -1; // one shot
		  pf_cmd = i2c_in_msg[1];
			pf_cmd <<= 8;
		  pf_cmd |= i2c_in_msg[2];
			pf_send_start();
		  break;
    default:
		  break;
	}
}

ISR(TWI_vect) 
{
	switch (TWSR) {
	  case 0x60: // received own slave address, master will write
		  i2c_in_msg_len = 0;
      TWCR |= _BV(TWEA);
		  break;
    case 0x80: // addressed as slave, data byte received, returned ACK
		  i2c_in_msg[ i2c_in_msg_len ] = TWDR;
      if (i2c_in_msg_len == 4 - 2)
        TWCR &= ~_BV(TWEA); // next one should be the last one
      else
        TWCR |= _BV(TWEA); // can receive more after next
			i2c_in_msg_len++;
			break;
    case 0x88: // addressed as slave, data byte received, returned NACK
		  i2c_in_msg[ i2c_in_msg_len ] = TWDR;
      TWCR |= _BV(TWEA); // back to idle slave mode
			i2c_process_in_msg();
			// read transmission ended
			break;
    case 0xA0: // stop or repeated start while addressed as slave
      TWCR |= _BV(TWEA); // back to idle slave mode
			i2c_process_in_msg();
			// read transmission ended
		  break;

    case 0xA8: // received own slave address, master will read
		  if (pf_cmd_recv_valid) {
			  i2c_out_msg[0] = (pf_cmd_recv & 0xff);
				i2c_out_msg[1] = (pf_cmd_recv >> 8);
				i2c_out_msg[2] = 1;
				i2c_out_msg_len = 3;
				pf_cmd_recv_valid = 0; // don't report same cmd twice
      } else {
			  i2c_out_msg[0] = 0;
				i2c_out_msg[1] = 0;
				i2c_out_msg[2] = 0;
				i2c_out_msg_len = 3;
			}
    case 0xB8: // byte sent, master responded with ACK
			if (i2c_out_msg_len == 0) { // error condition
			  TWDR = 0x00;
			} else {
		    i2c_out_msg_len--;
  			TWDR = i2c_out_msg[ i2c_out_msg_len - 1 ];
      }
			if (i2c_out_msg_len == 0) {
        TWCR &= ~_BV(TWEA); // last byte; expect NACK
      } else {
        TWCR |= _BV(TWEA); // more to send; expect ACK
			} 
		  break;
    case 0xC0: // data byte transmitted, master responded with NACK
    case 0xC8: // LAST byte transmitted (expected NACK), master responded with ACK
      TWCR |= _BV(TWEA); // just switch to idle slave mode
		  break;
    default:
		  // all other case should not occur in this system; error
			break;
	}
}

void i2c_slave_init() 
{
  uart_puts("i2c slave init\r\n");

  cli();

	TWAR = (SLAVE_ADDRESS << 1);

  TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA); // enable module, enable ints, enable ack

	sei();
}

/*
 * Automatic configuration of the I2C pull-ups
 */

void i2c_pullup_config()
{
  char     buffer[7];
  uint16_t result;

  PRR &= ~_BV(PRADC);   // make sure that the ADC is not shut down
  ADCSRA =   _BV(ADEN)  // enable the ADC
	         | _BV(ADPS0) | _BV(ADPS1) | _BV(ADPS2); // prescale by 128
					 // to get ADC clock of 50 to 200kHz (here from the 14.7MHz clock).
	 
  ADMUX =   _BV(REFS0) // use AVcc as a reference
	        | _BV(MUX0) | _BV(MUX2); // select ADC5=SCL

  ADCSRA |= _BV(ADSC); // start a conversion, single-conversion mode
	while (ADCSRA & _BV(ADSC)); // wait for completion

  /*
	DDRC |= _BV(DDC3);  // we use PC3 as the pull-up point (not VDD)
	DDRC &= ~_BV(DDC2); // PC2 also has an 82kOhm to SCL, put into high-z
  PORTC |= _BV(PORTC3); // pull up

  ADCSRA |= _BV(ADSC); // do it again, we discard the first conversion
	while (ADCSRA & _BV(ADSC)); // wait for completion

  result = ADC & 0x3FF;
  uart_puts_P("scl voltage with pullup = ");
  itoa(result, buffer, 10);
  uart_puts(buffer);     
  uart_puts_P("\r\n");

	DDRC  |= _BV(DDC3) | _BV(DDC2); // PC2 also has an 82kOhm to SCL
  PORTC |= _BV(PORTC2);  // pull up (simulate a pullup on another device)
  PORTC &= ~_BV(PORTC3); // pull down here

  ADCSRA |= _BV(ADSC); // do it again, we discard the first conversion
	while (ADCSRA & _BV(ADSC)); // wait for completion
	
  result = ADC & 0x3FF;
  uart_puts_P("scl voltage with pulldown+pulldown = ");
  itoa(result, buffer, 10);
  uart_puts(buffer);     
  uart_puts_P("\r\n");

	*/

	DDRC |= _BV(DDC3);  // we use PC3 as the pull-up point (not VDD)
	DDRC &= ~_BV(DDC2); // PC2 also has an 82kOhm to SCL, put into high-z
  PORTC &= ~_BV(PORTC3); // pull down

  ADCSRA |= _BV(ADSC); // do it again, we discard the first conversion
	while (ADCSRA & _BV(ADSC)); // wait for completion
	
  result = ADC & 0x3FF;
  uart_puts_P("scl voltage with pulldown = ");
  itoa(result, buffer, 10);
  uart_puts(buffer);     
  uart_puts_P("\r\n");

  if (result > 128) {
	  // we don't need the pullups, something else is pulling it up
    uart_puts_P("disabling local pullups on i2c lines\r\n");
  	DDRC &= ~_BV(DDC3); // put PC3 into high-z
	} else {
	  // enable pullups
    uart_puts_P("enabling local pullups on i2c lines\r\n");
  	DDRC |= _BV(DDC3);
    PORTC |= _BV(PORTC3); // pull up
  }
  
  ADCSRA = 0; // disable the ADC (ADEN is set to zero here)
  PRR |= _BV(PRADC);  // shut down the ADC

}

/*
 * Main routine.
 */

void main() {
	uint16_t i;

	uint16_t count = 0;


  unsigned int c;
  char buffer[7];
  int  num=134;

		int x = 0;
    
  /*
   *  Initialize UART library, pass baudrate and AVR cpu clock
   *  with the macro 
   *  UART_BAUD_SELECT() (normal speed mode )
   *  or 
   *  UART_BAUD_SELECT_DOUBLE_SPEED() ( double speed mode)
   */
  uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) ); 
    
  /*
   * now enable interrupt, since UART library is interrupt controlled
   */
  sei();
    
  /*
   *  Transmit string to UART
   *  The string is buffered by the uart library in a circular buffer
   *  and one character at a time is transmitted to the UART using interrupts.
   *  uart_puts() blocks if it can not write the whole string to the circular 
   *  buffer
   */
	
  uart_puts("String stored in SRAM\n");
    
  /*
   * Transmit string from program memory to UART
   */
  uart_puts_P("String stored in FLASH\n");
    
        
  /* 
   * Use standard avr-libc functions to convert numbers into string
   * before transmitting via UART
   */     
  itoa( num, buffer, 10);   // convert interger into string (decimal format)         
  uart_puts(buffer);        // and transmit string to UART

    
  /*
   * Transmit single character to UART
   */
  uart_putc('\r');
  uart_putc('\n');

  pf_init();
  //pf_send_init();
	i2c_pullup_config();
  i2c_slave_init();

	while (1);
	
	while (1) {
	  uint16_t l;
		uint8_t msg[4];
		cli();
 	  if (i2c_processed)
		  l = pf_cmd2;
			i2c_processed = 0;
		sei();
		if (l != 0) {
      uart_putc('*');
	    itoa(l, buffer, 10);
      uart_puts(buffer);     
      uart_putc('\r');
      uart_putc('\n');
			l = 0;
    }   
	}


  pf_receive();



	while (1) {
	  pf_cmd = 0xB141;
		pf_send_start();
		while (1) {
  		cli();
  	  if (state == rcv_done) {
 			  sei();
        break;
      }
 			sei();
	  }
		TCNT1 = 0;
		while (TCNT1 < 15000);
    uart_puts("***\r\n");
  	for (i=0; i<nperiods; i++) {
	    itoa(periods[i], buffer, 10);
      uart_puts(buffer);     
      uart_putc('\r');
      uart_putc('\n');
	  }
		//for (dummy=0; dummy < 300000; dummy++);
		//while (1);
	}

    
  for(;;)
  {
      /*
       * Get received character from ringbuffer
       * uart_getc() returns in the lower byte the received character and 
       * in the higher byte (bitmask) the last receive error
       * UART_NO_DATA is returned when no data is available.
       *
       */
      c = uart_getc();
      if ( c & UART_NO_DATA )
      {
          /* 
             * no data available from UART 
             */
        }
        else
        {
            /*
             * new data available from UART
             * check for Frame or Overrun error
             */
            if ( c & UART_FRAME_ERROR )
            {
                /* Framing Error detected, i.e no stop bit detected */
                uart_puts_P("UART Frame Error: ");
            }
            if ( c & UART_OVERRUN_ERROR )
            {
                /* 
                 * Overrun, a character already present in the UART UDR register was 
                 * not read by the interrupt handler before the next character arrived,
                 * one or more received characters have been dropped
                 */
                uart_puts_P("UART Overrun Error: ");
            }
            if ( c & UART_BUFFER_OVERFLOW )
            {
                /* 
                 * We are not reading the receive buffer fast enough,
                 * one or more received character have been dropped 
                 */
                uart_puts_P("Buffer overflow error: ");
            }
            /* 
             * send received character back
             */
            uart_putc( (unsigned char)c );
        }
    }

/*

	DDRA = 0xff; // output
  while (1) {

		PORTA |= 0x80;

		for (i=0; i<30000; i++) 
			if (dummy==1) PORTA = 0;

		PORTA &= 0x7f;

		for (i=0; i<30000; i++)
			if (dummy==1) PORTA = 0;

  }
	*/
}

#endif

// old scope view

#if 0

volatile uint8_t dummy = 0;
volatile uint8_t longquiet = 0;

volatile uint8_t quiet = 0;
volatile uint8_t  nperiods = 0;
volatile uint16_t periods[64];

volatile enum {
  idle = 0,
	low,
  high,
  done
} state; 

/*
ISR(TIMER1_OVF_vect) {
  switch (state) {
	  case idle:
		  quiet++;
      TCCR1B |= _BV(ICES1); // wait for rising edge, it case it messed up
		  break;
    case low:
			// this is really an error condition
			// with the detector stuck at active
			// low for a long time
	  case high:
			quiet = 1; // already overflowed once...
		  state = done;
		  break;
	  case done:
		  break;
	}
}
*/

ISR(TIMER1_COMPA_vect) {
  TCNT1 = 0;
  switch (state) {
	  case idle:
		  quiet++;
      //TCCR1B |= _BV(ICES1); // xxx wait for rising edge, it case it messed up
		  break;
    case low:
			// this is really an error condition
			// with the detector stuck at active
			// low for a long time
	  case high:
			quiet = 1; // already overflowed once...
		  state = done;
		  break;
	  case done:
		  break;
	}
}

ISR(TIMER1_CAPT_vect) {
  switch (state) {
	  case idle:
		  nperiods = 0;
      TCCR1B |= _BV(ICES1); // wait for rising edge
			state = low;
		  break;
    case low:
      periods[nperiods] = ICR1;
      nperiods++;
      TCCR1B &= ~(_BV(ICES1)); // wait for falling edge
			state = high;
		  break;
	  case high:
      periods[nperiods] = ICR1;
      nperiods++;
      TCCR1B |= _BV(ICES1); // wait for rising edge
			state = low;
		  break;
	  case done:
		  break;
	}

  TCNT1 = TCNT1 - ICR1;
	if (nperiods == 64) nperiods = 63; // overflow
}

void main() {
	uint16_t i;

	uint16_t count = 0;


  unsigned int c;
  char buffer[7];
  int  num=134;

    
  uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) ); 

 	TCCR1B =  _BV(ICNC1) // noise canceler
	                     // not setting ICES1, so falling edge
         | _BV(CS11) | _BV(CS10) // 011, meaning divide by 64
				 ; 

  //TIMSK1 = _BV(ICIE1)  // enable capture interrupt
	//       | _BV(TOIE1); // enable overflow interrupt

  TIMSK1 = _BV(ICIE1)   // enable capture interrupt
	       | _BV(TOIE1)   // enable overflow interrupt
	       | _BV(OCIE1A); // enable compare-match-A interrupt


  TCNT1 = 0;
  OCR1A = 768; // maximum high before we declare detection

	sei();

  state = idle;
	while (1) {
	  uint8_t print = 0;
		cli();
  	if (state == done) {
			print = 1;
		}
		sei();
		if (print) {
      uart_puts_P("d");
      itoa(count, buffer, 10);
      uart_puts(buffer);
      uart_puts(" = [ ... \r\n");
			count++;
			for (i=0; i<nperiods; i++) {
			    itoa(periods[i], buffer, 10);   // convert interger into string (decimal format)         
          uart_puts(buffer);        // and transmit string to UART
          uart_putc('\r');
          uart_putc('\n');
			}
      uart_puts("]\r\n");
			cli();
			state = idle;
			sei();
    }
	}
}

#endif
